﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq.Expressions;
using System.Windows;

namespace Common
{
    public class NotifySetterObject : INotifyPropertyChanged, INotifyPropertyChanging
    {
        public event PropertyChangedEventHandler PropertyChanged;
    
        public event PropertyChangingEventHandler PropertyChanging;

        protected void SetPropertyAndNotify<T>(ref T field, T value, params Expression<Func<object>>[] propertyExpression)
        {
            if (field == null ? value != null : !field.Equals(value))
            {
                foreach (var prop in propertyExpression)
                {
                    string propertyName = prop.ExtractPropertyName();
                    this.OnPropertyChanging(propertyName);
                }

                field = value;

                foreach (var prop in propertyExpression)
                {
                    string propertyName = prop.ExtractPropertyName();
                    this.OnPropertyChanged(propertyName);
                }
            }
        }

        protected void NotifyPropertyChanged<T>(Expression<Func<T>> propertyExpression)
        {
            this.OnPropertyChanged(propertyExpression.ExtractPropertyName());
        }
        
        private void OnPropertyChanging(string propertyName)
        {
            var handler = this.PropertyChanging;

            if (handler != null)
            {
                var action = new Action(() => handler(this, new PropertyChangingEventArgs(propertyName)));

                // determine if we need to switch back to the dispatcher thread no notify
                if (Application.Current.Dispatcher.CheckAccess())
                {
                    // currently on the dispatcher thread, no need to invoke across
                    action();
                }
                else
                {
                    // not on the dispatcher thread, queue an invocation with the dispatcher
                    Application.Current.Dispatcher.Invoke(action);
                }
            }
        }

        private void OnPropertyChanged(string propertyName)
        {
            var handler = this.PropertyChanged;

            if (handler != null)
            {
                var action = new Action(() =>
                {
                    try
                    {
                        handler(this, new PropertyChangedEventArgs(propertyName));
                    }
                    catch (Exception exception)
                    {
                        Debug.WriteLine(exception);
                    }
                });

                // determine if we need to switch back to the dispatcher thread no notify
                if (Application.Current.Dispatcher.CheckAccess())
                {
                    // currently on the dispatcher thread, no need to invoke across
                    action();
                }
                else
                {
                    // not on the dispatcher thread, queue an invocation with the dispatcher
                    Application.Current.Dispatcher.Invoke(action);
                }
            }
        }
    }
}
